@@ -18,6 +18,7 @@ package core
 
 import (
 	"fmt"
+	"k8s.io/autoscaler/cluster-autoscaler/core/utils"
 	"regexp"
 	"strings"
 	"testing"
@@ -28,7 +29,6 @@ import (
 	"k8s.io/autoscaler/cluster-autoscaler/clusterstate"
 	"k8s.io/autoscaler/cluster-autoscaler/config"
 	"k8s.io/autoscaler/cluster-autoscaler/estimator"
-	ca_processors "k8s.io/autoscaler/cluster-autoscaler/processors"
 	kube_util "k8s.io/autoscaler/cluster-autoscaler/utils/kubernetes"
 	. "k8s.io/autoscaler/cluster-autoscaler/utils/test"
 	"k8s.io/autoscaler/cluster-autoscaler/utils/units"
@@ -51,6 +51,7 @@ var defaultOptions = config.AutoscalingOptions{
 	MinMemoryTotal: 0,
 }
 
+// Scale up scenarios.
 func TestScaleUpOK(t *testing.T) {
 	config := &scaleTestConfig{
 		nodes: []nodeConfig{
@@ -64,12 +65,14 @@ func TestScaleUpOK(t *testing.T) {
 		extraPods: []podConfig{
 			{"p-new", 500, 0, 0, ""},
 		},
-		scaleUpOptionToChoose: groupSizeChange{groupName: "ng2", sizeChange: 1},
-		expectedFinalScaleUp:  groupSizeChange{groupName: "ng2", sizeChange: 1},
-		options:               defaultOptions,
+		options:                 defaultOptions,
+		expansionOptionToChoose: groupSizeChange{groupName: "ng2", sizeChange: 1},
+	}
+	expectedResults := &scaleTestResults{
+		finalOption: groupSizeChange{groupName: "ng2", sizeChange: 1},
 	}
 
-	simpleScaleUpTest(t, config)
+	simpleScaleUpTest(t, config, expectedResults)
 }
 
 func TestScaleUpMaxCoresLimitHit(t *testing.T) {
@@ -88,12 +91,14 @@ func TestScaleUpMaxCoresLimitHit(t *testing.T) {
 			{"p-new-1", 2000, 0, 0, ""},
 			{"p-new-2", 2000, 0, 0, ""},
 		},
-		scaleUpOptionToChoose: groupSizeChange{groupName: "ng1", sizeChange: 2},
-		expectedFinalScaleUp:  groupSizeChange{groupName: "ng1", sizeChange: 1},
-		options:               options,
+		expansionOptionToChoose: groupSizeChange{groupName: "ng1", sizeChange: 2},
+		options:                 options,
+	}
+	results := &scaleTestResults{
+		finalOption: groupSizeChange{groupName: "ng1", sizeChange: 1},
 	}
 
-	simpleScaleUpTest(t, config)
+	simpleScaleUpTest(t, config, results)
 }
 
 func TestScaleUpMaxCoresLimitHitWithNotAutoscaledGroup(t *testing.T) {
@@ -112,62 +117,68 @@ func TestScaleUpMaxCoresLimitHitWithNotAutoscaledGroup(t *testing.T) {
 			{"p-new-1", 2000, 0, 0, ""},
 			{"p-new-2", 2000, 0, 0, ""},
 		},
-		scaleUpOptionToChoose: groupSizeChange{groupName: "ng1", sizeChange: 2},
-		expectedFinalScaleUp:  groupSizeChange{groupName: "ng1", sizeChange: 1},
-		options:               options,
+		expansionOptionToChoose: groupSizeChange{groupName: "ng1", sizeChange: 2},
+		options:                 options,
+	}
+	results := &scaleTestResults{
+		finalOption: groupSizeChange{groupName: "ng1", sizeChange: 1},
 	}
 
-	simpleScaleUpTest(t, config)
+	simpleScaleUpTest(t, config, results)
 }
 
 func TestScaleUpMaxMemoryLimitHit(t *testing.T) {
 	options := defaultOptions
-	options.MaxMemoryTotal = 1300 * MiB
+	options.MaxMemoryTotal = 1300 * utils.MiB
 	config := &scaleTestConfig{
 		nodes: []nodeConfig{
-			{"n1", 2000, 100 * MiB, 0, true, "ng1"},
-			{"n2", 4000, 1000 * MiB, 0, true, "ng2"},
+			{"n1", 2000, 100 * utils.MiB, 0, true, "ng1"},
+			{"n2", 4000, 1000 * utils.MiB, 0, true, "ng2"},
 		},
 		pods: []podConfig{
 			{"p1", 1000, 0, 0, "n1"},
 			{"p2", 3000, 0, 0, "n2"},
 		},
 		extraPods: []podConfig{
-			{"p-new-1", 2000, 100 * MiB, 0, ""},
-			{"p-new-2", 2000, 100 * MiB, 0, ""},
-			{"p-new-3", 2000, 100 * MiB, 0, ""},
+			{"p-new-1", 2000, 100 * utils.MiB, 0, ""},
+			{"p-new-2", 2000, 100 * utils.MiB, 0, ""},
+			{"p-new-3", 2000, 100 * utils.MiB, 0, ""},
 		},
-		scaleUpOptionToChoose: groupSizeChange{groupName: "ng1", sizeChange: 3},
-		expectedFinalScaleUp:  groupSizeChange{groupName: "ng1", sizeChange: 2},
-		options:               options,
+		expansionOptionToChoose: groupSizeChange{groupName: "ng1", sizeChange: 3},
+		options:                 options,
+	}
+	results := &scaleTestResults{
+		finalOption: groupSizeChange{groupName: "ng1", sizeChange: 2},
 	}
 
-	simpleScaleUpTest(t, config)
+	simpleScaleUpTest(t, config, results)
 }
 
 func TestScaleUpMaxMemoryLimitHitWithNotAutoscaledGroup(t *testing.T) {
 	options := defaultOptions
-	options.MaxMemoryTotal = 1300 * MiB
+	options.MaxMemoryTotal = 1300 * utils.MiB
 	config := &scaleTestConfig{
 		nodes: []nodeConfig{
-			{"n1", 2000, 100 * MiB, 0, true, "ng1"},
-			{"n2", 4000, 1000 * MiB, 0, true, ""},
+			{"n1", 2000, 100 * utils.MiB, 0, true, "ng1"},
+			{"n2", 4000, 1000 * utils.MiB, 0, true, ""},
 		},
 		pods: []podConfig{
 			{"p1", 1000, 0, 0, "n1"},
 			{"p2", 3000, 0, 0, "n2"},
 		},
 		extraPods: []podConfig{
-			{"p-new-1", 2000, 100 * MiB, 0, ""},
-			{"p-new-2", 2000, 100 * MiB, 0, ""},
-			{"p-new-3", 2000, 100 * MiB, 0, ""},
+			{"p-new-1", 2000, 100 * utils.MiB, 0, ""},
+			{"p-new-2", 2000, 100 * utils.MiB, 0, ""},
+			{"p-new-3", 2000, 100 * utils.MiB, 0, ""},
 		},
-		scaleUpOptionToChoose: groupSizeChange{groupName: "ng1", sizeChange: 3},
-		expectedFinalScaleUp:  groupSizeChange{groupName: "ng1", sizeChange: 2},
-		options:               options,
+		expansionOptionToChoose: groupSizeChange{groupName: "ng1", sizeChange: 3},
+		options:                 options,
+	}
+	results := &scaleTestResults{
+		finalOption: groupSizeChange{groupName: "ng1", sizeChange: 2},
 	}
 
-	simpleScaleUpTest(t, config)
+	simpleScaleUpTest(t, config, results)
 }
 
 func TestScaleUpCapToMaxTotalNodesLimit(t *testing.T) {
@@ -175,24 +186,26 @@ func TestScaleUpCapToMaxTotalNodesLimit(t *testing.T) {
 	options.MaxNodesTotal = 3
 	config := &scaleTestConfig{
 		nodes: []nodeConfig{
-			{"n1", 2000, 100 * MiB, 0, true, "ng1"},
-			{"n2", 4000, 1000 * MiB, 0, true, "ng2"},
+			{"n1", 2000, 100 * utils.MiB, 0, true, "ng1"},
+			{"n2", 4000, 1000 * utils.MiB, 0, true, "ng2"},
 		},
 		pods: []podConfig{
 			{"p1", 1000, 0, 0, "n1"},
 			{"p2", 3000, 0, 0, "n2"},
 		},
 		extraPods: []podConfig{
-			{"p-new-1", 4000, 100 * MiB, 0, ""},
-			{"p-new-2", 4000, 100 * MiB, 0, ""},
-			{"p-new-3", 4000, 100 * MiB, 0, ""},
+			{"p-new-1", 4000, 100 * utils.MiB, 0, ""},
+			{"p-new-2", 4000, 100 * utils.MiB, 0, ""},
+			{"p-new-3", 4000, 100 * utils.MiB, 0, ""},
 		},
-		scaleUpOptionToChoose: groupSizeChange{groupName: "ng2", sizeChange: 3},
-		expectedFinalScaleUp:  groupSizeChange{groupName: "ng2", sizeChange: 1},
-		options:               options,
+		expansionOptionToChoose: groupSizeChange{groupName: "ng2", sizeChange: 3},
+		options:                 options,
+	}
+	results := &scaleTestResults{
+		finalOption: groupSizeChange{groupName: "ng2", sizeChange: 1},
 	}
 
-	simpleScaleUpTest(t, config)
+	simpleScaleUpTest(t, config, results)
 }
 
 func TestScaleUpCapToMaxTotalNodesLimitWithNotAutoscaledGroup(t *testing.T) {
@@ -200,24 +213,26 @@ func TestScaleUpCapToMaxTotalNodesLimitWithNotAutoscaledGroup(t *testing.T) {
 	options.MaxNodesTotal = 3
 	config := &scaleTestConfig{
 		nodes: []nodeConfig{
-			{"n1", 2000, 100 * MiB, 0, true, ""},
-			{"n2", 4000, 1000 * MiB, 0, true, "ng2"},
+			{"n1", 2000, 100 * utils.MiB, 0, true, ""},
+			{"n2", 4000, 1000 * utils.MiB, 0, true, "ng2"},
 		},
 		pods: []podConfig{
 			{"p1", 1000, 0, 0, "n1"},
 			{"p2", 3000, 0, 0, "n2"},
 		},
 		extraPods: []podConfig{
-			{"p-new-1", 4000, 100 * MiB, 0, ""},
-			{"p-new-2", 4000, 100 * MiB, 0, ""},
-			{"p-new-3", 4000, 100 * MiB, 0, ""},
+			{"p-new-1", 4000, 100 * utils.MiB, 0, ""},
+			{"p-new-2", 4000, 100 * utils.MiB, 0, ""},
+			{"p-new-3", 4000, 100 * utils.MiB, 0, ""},
 		},
-		scaleUpOptionToChoose: groupSizeChange{groupName: "ng2", sizeChange: 3},
-		expectedFinalScaleUp:  groupSizeChange{groupName: "ng2", sizeChange: 1},
-		options:               options,
+		expansionOptionToChoose: groupSizeChange{groupName: "ng2", sizeChange: 3},
+		options:                 options,
+	}
+	results := &scaleTestResults{
+		finalOption: groupSizeChange{groupName: "ng2", sizeChange: 1},
 	}
 
-	simpleScaleUpTest(t, config)
+	simpleScaleUpTest(t, config, results)
 }
 
 func TestWillConsiderGpuAndStandardPoolForPodWhichDoesNotRequireGpu(t *testing.T) {
@@ -225,26 +240,28 @@ func TestWillConsiderGpuAndStandardPoolForPodWhichDoesNotRequireGpu(t *testing.T
 	options.MaxNodesTotal = 100
 	config := &scaleTestConfig{
 		nodes: []nodeConfig{
-			{"gpu-node-1", 2000, 1000 * MiB, 1, true, "gpu-pool"},
-			{"std-node-1", 2000, 1000 * MiB, 0, true, "std-pool"},
+			{"gpu-node-1", 2000, 1000 * utils.MiB, 1, true, "gpu-pool"},
+			{"std-node-1", 2000, 1000 * utils.MiB, 0, true, "std-pool"},
 		},
 		pods: []podConfig{
-			{"gpu-pod-1", 2000, 1000 * MiB, 1, "gpu-node-1"},
-			{"std-pod-1", 2000, 1000 * MiB, 0, "std-node-1"},
+			{"gpu-pod-1", 2000, 1000 * utils.MiB, 1, "gpu-node-1"},
+			{"std-pod-1", 2000, 1000 * utils.MiB, 0, "std-node-1"},
 		},
 		extraPods: []podConfig{
-			{"extra-std-pod", 2000, 1000 * MiB, 0, ""},
+			{"extra-std-pod", 2000, 1000 * utils.MiB, 0, ""},
 		},
-		expectedScaleUpOptions: []groupSizeChange{
+		expansionOptionToChoose: groupSizeChange{groupName: "std-pool", sizeChange: 1},
+		options:                 options,
+	}
+	results := &scaleTestResults{
+		finalOption: groupSizeChange{groupName: "std-pool", sizeChange: 1},
+		expansionOptions: []groupSizeChange{
 			{groupName: "std-pool", sizeChange: 1},
 			{groupName: "gpu-pool", sizeChange: 1},
 		},
-		scaleUpOptionToChoose: groupSizeChange{groupName: "std-pool", sizeChange: 1},
-		expectedFinalScaleUp:  groupSizeChange{groupName: "std-pool", sizeChange: 1},
-		options:               options,
 	}
 
-	simpleScaleUpTest(t, config)
+	simpleScaleUpTest(t, config, results)
 }
 
 func TestWillConsiderOnlyGpuPoolForPodWhichDoesRequiresGpu(t *testing.T) {
@@ -252,25 +269,27 @@ func TestWillConsiderOnlyGpuPoolForPodWhichDoesRequiresGpu(t *testing.T) {
 	options.MaxNodesTotal = 100
 	config := &scaleTestConfig{
 		nodes: []nodeConfig{
-			{"gpu-node-1", 2000, 1000 * MiB, 1, true, "gpu-pool"},
-			{"std-node-1", 2000, 1000 * MiB, 0, true, "std-pool"},
+			{"gpu-node-1", 2000, 1000 * utils.MiB, 1, true, "gpu-pool"},
+			{"std-node-1", 2000, 1000 * utils.MiB, 0, true, "std-pool"},
 		},
 		pods: []podConfig{
-			{"gpu-pod-1", 2000, 1000 * MiB, 1, "gpu-node-1"},
-			{"std-pod-1", 2000, 1000 * MiB, 0, "std-node-1"},
+			{"gpu-pod-1", 2000, 1000 * utils.MiB, 1, "gpu-node-1"},
+			{"std-pod-1", 2000, 1000 * utils.MiB, 0, "std-node-1"},
 		},
 		extraPods: []podConfig{
-			{"extra-gpu-pod", 2000, 1000 * MiB, 1, ""},
+			{"extra-gpu-pod", 2000, 1000 * utils.MiB, 1, ""},
 		},
-		expectedScaleUpOptions: []groupSizeChange{
+		expansionOptionToChoose: groupSizeChange{groupName: "gpu-pool", sizeChange: 1},
+		options:                 options,
+	}
+	results := &scaleTestResults{
+		finalOption: groupSizeChange{groupName: "gpu-pool", sizeChange: 1},
+		expansionOptions: []groupSizeChange{
 			{groupName: "gpu-pool", sizeChange: 1},
 		},
-		scaleUpOptionToChoose: groupSizeChange{groupName: "gpu-pool", sizeChange: 1},
-		expectedFinalScaleUp:  groupSizeChange{groupName: "gpu-pool", sizeChange: 1},
-		options:               options,
 	}
 
-	simpleScaleUpTest(t, config)
+	simpleScaleUpTest(t, config, results)
 }
 
 func TestWillConsiderAllPoolsWhichFitTwoPodsRequiringGpus(t *testing.T) {
@@ -278,67 +297,87 @@ func TestWillConsiderAllPoolsWhichFitTwoPodsRequiringGpus(t *testing.T) {
 	options.MaxNodesTotal = 100
 	config := &scaleTestConfig{
 		nodes: []nodeConfig{
-			{"gpu-1-node-1", 2000, 1000 * MiB, 1, true, "gpu-1-pool"},
-			{"gpu-2-node-1", 2000, 1000 * MiB, 2, true, "gpu-2-pool"},
-			{"gpu-4-node-1", 2000, 1000 * MiB, 4, true, "gpu-4-pool"},
-			{"std-node-1", 2000, 1000 * MiB, 0, true, "std-pool"},
+			{"gpu-1-node-1", 2000, 1000 * utils.MiB, 1, true, "gpu-1-pool"},
+			{"gpu-2-node-1", 2000, 1000 * utils.MiB, 2, true, "gpu-2-pool"},
+			{"gpu-4-node-1", 2000, 1000 * utils.MiB, 4, true, "gpu-4-pool"},
+			{"std-node-1", 2000, 1000 * utils.MiB, 0, true, "std-pool"},
 		},
 		pods: []podConfig{
-			{"gpu-pod-1", 2000, 1000 * MiB, 1, "gpu-1-node-1"},
-			{"gpu-pod-2", 2000, 1000 * MiB, 2, "gpu-2-node-1"},
-			{"gpu-pod-3", 2000, 1000 * MiB, 4, "gpu-4-node-1"},
-			{"std-pod-1", 2000, 1000 * MiB, 0, "std-node-1"},
+			{"gpu-pod-1", 2000, 1000 * utils.MiB, 1, "gpu-1-node-1"},
+			{"gpu-pod-2", 2000, 1000 * utils.MiB, 2, "gpu-2-node-1"},
+			{"gpu-pod-3", 2000, 1000 * utils.MiB, 4, "gpu-4-node-1"},
+			{"std-pod-1", 2000, 1000 * utils.MiB, 0, "std-node-1"},
 		},
 		extraPods: []podConfig{
-			{"extra-gpu-pod-1", 1, 1 * MiB, 1, ""}, // CPU and mem negligible
-			{"extra-gpu-pod-2", 1, 1 * MiB, 1, ""}, // CPU and mem negligible
-			{"extra-gpu-pod-3", 1, 1 * MiB, 1, ""}, // CPU and mem negligible
+			{"extra-gpu-pod-1", 1, 1 * utils.MiB, 1, ""}, // CPU and mem negligible
+			{"extra-gpu-pod-2", 1, 1 * utils.MiB, 1, ""}, // CPU and mem negligible
+			{"extra-gpu-pod-3", 1, 1 * utils.MiB, 1, ""}, // CPU and mem negligible
 		},
-		expectedScaleUpOptions: []groupSizeChange{
+		expansionOptionToChoose: groupSizeChange{groupName: "gpu-1-pool", sizeChange: 3},
+		options:                 options,
+	}
+	results := &scaleTestResults{
+		finalOption: groupSizeChange{groupName: "gpu-1-pool", sizeChange: 3},
+		expansionOptions: []groupSizeChange{
 			{groupName: "gpu-1-pool", sizeChange: 3},
 			{groupName: "gpu-2-pool", sizeChange: 2},
 			{groupName: "gpu-4-pool", sizeChange: 1},
 		},
-		scaleUpOptionToChoose: groupSizeChange{groupName: "gpu-1-pool", sizeChange: 3},
-		expectedFinalScaleUp:  groupSizeChange{groupName: "gpu-1-pool", sizeChange: 3},
-		options:               options,
 	}
 
-	simpleScaleUpTest(t, config)
+	simpleScaleUpTest(t, config, results)
 }
 
-type assertingStrategy struct {
-	initialNodeConfigs     []nodeConfig
-	expectedScaleUpOptions []groupSizeChange
-	scaleUpOptionToChoose  groupSizeChange
-	t                      *testing.T
-}
-
-func (s assertingStrategy) BestOption(options []expander.Option, nodeInfo map[string]*schedulernodeinfo.NodeInfo) *expander.Option {
-	if len(s.expectedScaleUpOptions) > 0 {
-		// empty s.expectedScaleUpOptions means we do not want to do assertion on contents of actual scaleUp options
-
-		// precondition check that option to choose is part of expected options
-		assert.Contains(s.t, s.expectedScaleUpOptions, s.scaleUpOptionToChoose, "scaleUpOptionToChoose must be present in expectedScaleUpOptions")
-
-		actualScaleUpOptions := expanderOptionsToGroupSizeChanges(options)
-		assert.Subset(s.t, actualScaleUpOptions, s.expectedScaleUpOptions,
-			"actual %s and expected %s scaleUp options differ",
-			actualScaleUpOptions,
-			s.expectedScaleUpOptions)
-		assert.Equal(s.t, len(actualScaleUpOptions), len(s.expectedScaleUpOptions),
-			"actual %s and expected %s scaleUp options differ",
-			actualScaleUpOptions,
-			s.expectedScaleUpOptions)
+// No scale up scenarios.
+func TestNoScaleUpMaxCoresLimitHit(t *testing.T) {
+	options := defaultOptions
+	options.MaxCoresTotal = 7
+	options.MaxMemoryTotal = 1150
+	config := &scaleTestConfig{
+		nodes: []nodeConfig{
+			{"n1", 2000, 100, 0, true, "ng1"},
+			{"n2", 4000, 1000, 0, true, "ng2"},
+		},
+		pods: []podConfig{
+			{"p1", 1000, 0, 0, "n1"},
+			{"p2", 3000, 0, 0, "n2"},
+		},
+		extraPods: []podConfig{
+			{"p-new-1", 2000, 0, 0, ""},
+			{"p-new-2", 2000, 0, 0, ""},
+		},
+		options: options,
+	}
+	results := &scaleTestResults{
+		noScaleUpReason: "max cluster cpu, memory limit reached",
 	}
 
+	simpleNoScaleUpTest(t, config, results)
+}
+
+// To implement expander.Strategy, BestOption method must have a struct receiver.
+// This prevents it from modifying fields of reportingStrategy, so we need a thin
+// pointer wrapper for mutable parts.
+type expanderResults struct {
+	inputOptions []groupSizeChange
+}
+
+type reportingStrategy struct {
+	initialNodeConfigs []nodeConfig
+	optionToChoose     groupSizeChange
+	results            *expanderResults
+	t                  *testing.T
+}
+
+func (r reportingStrategy) BestOption(options []expander.Option, nodeInfo map[string]*schedulernodeinfo.NodeInfo) *expander.Option {
+	r.results.inputOptions = expanderOptionsToGroupSizeChanges(options)
 	for _, option := range options {
-		scaleUpOption := expanderOptionToGroupSizeChange(option)
-		if scaleUpOption == s.scaleUpOptionToChoose {
+		groupSizeChange := expanderOptionToGroupSizeChange(option)
+		if groupSizeChange == r.optionToChoose {
 			return &option
 		}
 	}
-	assert.Fail(s.t, "did not find scaleUpOptionToChoose %s", s.scaleUpOptionToChoose)
+	assert.Fail(r.t, "did not find expansionOptionToChoose %s", r.optionToChoose)
 	return nil
 }
 
@@ -358,7 +397,7 @@ func expanderOptionToGroupSizeChange(option expander.Option) groupSizeChange {
 	return scaleUpOption
 }
 
-func simpleScaleUpTest(t *testing.T, config *scaleTestConfig) {
+func runSimpleScaleUpTest(t *testing.T, config *scaleTestConfig) *scaleTestResults {
 	expandedGroups := make(chan groupSizeChange, 10)
 
 	groups := make(map[string][]*apiv1.Node)
@@ -404,16 +443,16 @@ func simpleScaleUpTest(t *testing.T, config *scaleTestConfig) {
 	assert.NotNil(t, provider)
 
 	// Create context with non-random expander strategy.
-	context := NewScaleTestAutoscalingContext(config.options, &fake.Clientset{}, listers, provider)
-	expander := assertingStrategy{
-		initialNodeConfigs:     config.nodes,
-		expectedScaleUpOptions: config.expectedScaleUpOptions,
-		scaleUpOptionToChoose:  config.scaleUpOptionToChoose,
-		t:                      t,
+	context := NewScaleTestAutoscalingContext(config.options, &fake.Clientset{}, listers, provider, nil)
+	expander := reportingStrategy{
+		initialNodeConfigs: config.nodes,
+		optionToChoose:     config.expansionOptionToChoose,
+		results:            &expanderResults{},
+		t:                  t,
 	}
 	context.ExpanderStrategy = expander
 
-	nodeInfos, _ := getNodeInfosForGroups(nodes, nil, provider, listers, []*appsv1.DaemonSet{}, context.PredicateChecker)
+	nodeInfos, _ := utils.GetNodeInfosForGroups(nodes, nil, provider, listers, []*appsv1.DaemonSet{}, context.PredicateChecker, nil)
 	clusterState := clusterstate.NewClusterStateRegistry(provider, clusterstate.ClusterStateRegistryConfig{}, context.LogRecorder, newBackoff())
 	clusterState.UpdateNodes(nodes, nodeInfos, time.Now())
 
@@ -423,37 +462,94 @@ func simpleScaleUpTest(t *testing.T, config *scaleTestConfig) {
 		extraPods[i] = pod
 	}
 
-	processors := ca_processors.TestProcessors()
+	processors := NewTestProcessors()
 
-	scaleUpStatus, err := ScaleUp(&context, processors, clusterState, extraPods, nodes, []*appsv1.DaemonSet{}, nodeInfos)
+	scaleUpStatus, err := ScaleUp(&context, processors, clusterState, extraPods, nodes, []*appsv1.DaemonSet{}, nodeInfos, nil)
 	processors.ScaleUpStatusProcessor.Process(&context, scaleUpStatus)
+
 	assert.NoError(t, err)
-	assert.True(t, scaleUpStatus.WasSuccessful())
 
 	expandedGroup := getGroupSizeChangeFromChan(expandedGroups)
-	assert.NotNil(t, expandedGroup, "Expected scale up event")
-	assert.Equal(t, config.expectedFinalScaleUp, *expandedGroup)
+	var expandedGroupStruct groupSizeChange
+	if expandedGroup != nil {
+		expandedGroupStruct = *expandedGroup
+	}
 
-	nodeEventSeen := false
+	events := []string{}
 	for eventsLeft := true; eventsLeft; {
 		select {
 		case event := <-context.Recorder.(*kube_record.FakeRecorder).Events:
-			if strings.Contains(event, "TriggeredScaleUp") && strings.Contains(event, config.expectedFinalScaleUp.groupName) {
-				nodeEventSeen = true
-			}
-			assert.NotRegexp(t, regexp.MustCompile("NotTriggerScaleUp"), event)
+			events = append(events, event)
 		default:
 			eventsLeft = false
 		}
 	}
+
+	return &scaleTestResults{
+		expansionOptions: expander.results.inputOptions,
+		finalOption:      expandedGroupStruct,
+		scaleUpStatus:    scaleUpStatus,
+		events:           events,
+	}
+}
+
+func simpleNoScaleUpTest(t *testing.T, config *scaleTestConfig, expectedResults *scaleTestResults) {
+	results := runSimpleScaleUpTest(t, config)
+
+	assert.Equal(t, groupSizeChange{}, results.finalOption)
+	assert.False(t, results.scaleUpStatus.WasSuccessful())
+	noScaleUpEventSeen := false
+	for _, event := range results.events {
+		if strings.Contains(event, "NotTriggerScaleUp") {
+			if strings.Contains(event, expectedResults.noScaleUpReason) {
+				noScaleUpEventSeen = true
+			} else {
+				// Surprisingly useful for debugging.
+				fmt.Println("Event:", event)
+			}
+		}
+		assert.NotRegexp(t, regexp.MustCompile("TriggeredScaleUp"), event)
+	}
+	assert.True(t, noScaleUpEventSeen)
+}
+
+func simpleScaleUpTest(t *testing.T, config *scaleTestConfig, expectedResults *scaleTestResults) {
+	results := runSimpleScaleUpTest(t, config)
+
+	assert.NotNil(t, results.finalOption, "Expected scale up event")
+	assert.Equal(t, expectedResults.finalOption, results.finalOption)
+	assert.True(t, results.scaleUpStatus.WasSuccessful())
+	nodeEventSeen := false
+	for _, event := range results.events {
+		if strings.Contains(event, "TriggeredScaleUp") && strings.Contains(event, expectedResults.finalOption.groupName) {
+			nodeEventSeen = true
+		}
+		assert.NotRegexp(t, regexp.MustCompile("NotTriggerScaleUp"), event)
+	}
 	assert.True(t, nodeEventSeen)
+
+	if len(expectedResults.expansionOptions) > 0 {
+		// Empty expansionOptions means we do not want to do any assertions
+		// on contents of actual scaleUp options
+
+		// Check that option to choose is part of expected options.
+		assert.Contains(t, expectedResults.expansionOptions, config.expansionOptionToChoose, "final expected expansion option must be in expected expansion options")
+		assert.Contains(t, results.expansionOptions, config.expansionOptionToChoose, "final expected expansion option must be in expected expansion options")
+
+		assert.Subset(t, results.expansionOptions, expectedResults.expansionOptions,
+			"actual and expected expansion options should be the same")
+		assert.Subset(t, expectedResults.expansionOptions, results.expansionOptions,
+			"actual and expected expansion options should be the same")
+		assert.Equal(t, len(expectedResults.expansionOptions), len(results.expansionOptions),
+			"actual and expected expansion options should be the same")
+	}
 }
 
 func getGroupSizeChangeFromChan(c chan groupSizeChange) *groupSizeChange {
 	select {
 	case val := <-c:
 		return &val
-	case <-time.After(10 * time.Second):
+	case <-time.After(100 * time.Millisecond):
 		return nil
 	}
 }
@@ -469,106 +565,6 @@ func buildTestPod(p podConfig) *apiv1.Pod {
 	return pod
 }
 
-func TestScaleUpNodeComingNoScale(t *testing.T) {
-	n1 := BuildTestNode("n1", 100, 1000)
-	SetNodeReadyState(n1, true, time.Now())
-	n2 := BuildTestNode("n2", 1000, 1000)
-	SetNodeReadyState(n2, true, time.Now())
-
-	p1 := BuildTestPod("p1", 80, 0)
-	p2 := BuildTestPod("p2", 800, 0)
-	p1.Spec.NodeName = "n1"
-	p2.Spec.NodeName = "n2"
-
-	podLister := kube_util.NewTestPodLister([]*apiv1.Pod{p1, p2})
-	listers := kube_util.NewListerRegistry(nil, nil, podLister, nil, nil, nil, nil, nil, nil, nil)
-
-	provider := testprovider.NewTestCloudProvider(func(nodeGroup string, increase int) error {
-		t.Fatalf("No expansion is expected, but increased %s by %d", nodeGroup, increase)
-		return nil
-	}, nil)
-	provider.AddNodeGroup("ng1", 1, 10, 1)
-	provider.AddNodeGroup("ng2", 1, 10, 2)
-	provider.AddNode("ng1", n1)
-	provider.AddNode("ng2", n2)
-
-	options := config.AutoscalingOptions{
-		EstimatorName:  estimator.BinpackingEstimatorName,
-		MaxCoresTotal:  config.DefaultMaxClusterCores,
-		MaxMemoryTotal: config.DefaultMaxClusterMemory,
-	}
-	context := NewScaleTestAutoscalingContext(options, &fake.Clientset{}, listers, provider)
-
-	nodes := []*apiv1.Node{n1, n2}
-	nodeInfos, _ := getNodeInfosForGroups(nodes, nil, provider, listers, []*appsv1.DaemonSet{}, context.PredicateChecker)
-	clusterState := clusterstate.NewClusterStateRegistry(
-		provider,
-		clusterstate.ClusterStateRegistryConfig{MaxNodeProvisionTime: 5 * time.Minute},
-		context.LogRecorder,
-		newBackoff())
-	clusterState.RegisterOrUpdateScaleUp(provider.GetNodeGroup("ng2"), 1, time.Now())
-	clusterState.UpdateNodes(nodes, nodeInfos, time.Now())
-
-	p3 := BuildTestPod("p-new", 550, 0)
-
-	processors := ca_processors.TestProcessors()
-
-	scaleUpStatus, err := ScaleUp(&context, processors, clusterState, []*apiv1.Pod{p3}, nodes, []*appsv1.DaemonSet{}, nodeInfos)
-	assert.NoError(t, err)
-	// A node is already coming - no need for scale up.
-	assert.False(t, scaleUpStatus.WasSuccessful())
-}
-
-func TestScaleUpNodeComingHasScale(t *testing.T) {
-	n1 := BuildTestNode("n1", 100, 1000)
-	SetNodeReadyState(n1, true, time.Now())
-	n2 := BuildTestNode("n2", 1000, 1000)
-	SetNodeReadyState(n2, true, time.Now())
-
-	p1 := BuildTestPod("p1", 80, 0)
-	p2 := BuildTestPod("p2", 800, 0)
-	p1.Spec.NodeName = "n1"
-	p2.Spec.NodeName = "n2"
-
-	podLister := kube_util.NewTestPodLister([]*apiv1.Pod{p1, p2})
-	listers := kube_util.NewListerRegistry(nil, nil, podLister, nil, nil, nil, nil, nil, nil, nil)
-
-	expandedGroups := make(chan string, 10)
-	provider := testprovider.NewTestCloudProvider(func(nodeGroup string, increase int) error {
-		expandedGroups <- fmt.Sprintf("%s-%d", nodeGroup, increase)
-		return nil
-	}, nil)
-	provider.AddNodeGroup("ng1", 1, 10, 1)
-	provider.AddNodeGroup("ng2", 1, 10, 2)
-	provider.AddNode("ng1", n1)
-	provider.AddNode("ng2", n2)
-
-	context := NewScaleTestAutoscalingContext(defaultOptions, &fake.Clientset{}, listers, provider)
-
-	nodes := []*apiv1.Node{n1, n2}
-	nodeInfos, _ := getNodeInfosForGroups(nodes, nil, provider, listers, []*appsv1.DaemonSet{}, context.PredicateChecker)
-	clusterState := clusterstate.NewClusterStateRegistry(
-		provider,
-		clusterstate.ClusterStateRegistryConfig{
-			MaxNodeProvisionTime: 5 * time.Minute,
-		},
-		context.LogRecorder,
-		newBackoff())
-	clusterState.RegisterOrUpdateScaleUp(provider.GetNodeGroup("ng2"), 1, time.Now())
-	clusterState.UpdateNodes(nodes, nodeInfos, time.Now())
-
-	p3 := BuildTestPod("p-new", 550, 0)
-	p4 := BuildTestPod("p-new", 550, 0)
-
-	processors := ca_processors.TestProcessors()
-	scaleUpStatus, err := ScaleUp(&context, processors, clusterState, []*apiv1.Pod{p3, p4}, nodes, []*appsv1.DaemonSet{}, nodeInfos)
-
-	assert.NoError(t, err)
-	// Two nodes needed but one node is already coming, so it should increase by one.
-	assert.True(t, scaleUpStatus.WasSuccessful())
-	assert.Equal(t, "ng2-1", getStringFromChan(expandedGroups))
-}
-
 func TestScaleUpUnhealthy(t *testing.T) {
 	n1 := BuildTestNode("n1", 100, 1000)
 	SetNodeReadyState(n1, true, time.Now())
@@ -597,16 +593,16 @@ func TestScaleUpUnhealthy(t *testing.T) {
 		MaxCoresTotal:  config.DefaultMaxClusterCores,
 		MaxMemoryTotal: config.DefaultMaxClusterMemory,
 	}
-	context := NewScaleTestAutoscalingContext(options, &fake.Clientset{}, listers, provider)
+	context := NewScaleTestAutoscalingContext(options, &fake.Clientset{}, listers, provider, nil)
 
 	nodes := []*apiv1.Node{n1, n2}
-	nodeInfos, _ := getNodeInfosForGroups(nodes, nil, provider, listers, []*appsv1.DaemonSet{}, context.PredicateChecker)
+	nodeInfos, _ := utils.GetNodeInfosForGroups(nodes, nil, provider, listers, []*appsv1.DaemonSet{}, context.PredicateChecker, nil)
 	clusterState := clusterstate.NewClusterStateRegistry(provider, clusterstate.ClusterStateRegistryConfig{}, context.LogRecorder, newBackoff())
 	clusterState.UpdateNodes(nodes, nodeInfos, time.Now())
 	p3 := BuildTestPod("p-new", 550, 0)
 
-	processors := ca_processors.TestProcessors()
-	scaleUpStatus, err := ScaleUp(&context, processors, clusterState, []*apiv1.Pod{p3}, nodes, []*appsv1.DaemonSet{}, nodeInfos)
+	processors := NewTestProcessors()
+	scaleUpStatus, err := ScaleUp(&context, processors, clusterState, []*apiv1.Pod{p3}, nodes, []*appsv1.DaemonSet{}, nodeInfos, nil)
 
 	assert.NoError(t, err)
 	// Node group is unhealthy.
@@ -636,16 +632,16 @@ func TestScaleUpNoHelp(t *testing.T) {
 		MaxCoresTotal:  config.DefaultMaxClusterCores,
 		MaxMemoryTotal: config.DefaultMaxClusterMemory,
 	}
-	context := NewScaleTestAutoscalingContext(options, &fake.Clientset{}, listers, provider)
+	context := NewScaleTestAutoscalingContext(options, &fake.Clientset{}, listers, provider, nil)
 
 	nodes := []*apiv1.Node{n1}
-	nodeInfos, _ := getNodeInfosForGroups(nodes, nil, provider, listers, []*appsv1.DaemonSet{}, context.PredicateChecker)
+	nodeInfos, _ := utils.GetNodeInfosForGroups(nodes, nil, provider, listers, []*appsv1.DaemonSet{}, context.PredicateChecker, nil)
 	clusterState := clusterstate.NewClusterStateRegistry(provider, clusterstate.ClusterStateRegistryConfig{}, context.LogRecorder, newBackoff())
 	clusterState.UpdateNodes(nodes, nodeInfos, time.Now())
 	p3 := BuildTestPod("p-new", 500, 0)
 
-	processors := ca_processors.TestProcessors()
-	scaleUpStatus, err := ScaleUp(&context, processors, clusterState, []*apiv1.Pod{p3}, nodes, []*appsv1.DaemonSet{}, nodeInfos)
+	processors := NewTestProcessors()
+	scaleUpStatus, err := ScaleUp(&context, processors, clusterState, []*apiv1.Pod{p3}, nodes, []*appsv1.DaemonSet{}, nodeInfos, nil)
 	processors.ScaleUpStatusProcessor.Process(&context, scaleUpStatus)
 
 	assert.NoError(t, err)
@@ -701,9 +697,9 @@ func TestScaleUpBalanceGroups(t *testing.T) {
 		MaxCoresTotal:            config.DefaultMaxClusterCores,
 		MaxMemoryTotal:           config.DefaultMaxClusterMemory,
 	}
-	context := NewScaleTestAutoscalingContext(options, &fake.Clientset{}, listers, provider)
+	context := NewScaleTestAutoscalingContext(options, &fake.Clientset{}, listers, provider, nil)
 
-	nodeInfos, _ := getNodeInfosForGroups(nodes, nil, provider, listers, []*appsv1.DaemonSet{}, context.PredicateChecker)
+	nodeInfos, _ := utils.GetNodeInfosForGroups(nodes, nil, provider, listers, []*appsv1.DaemonSet{}, context.PredicateChecker, nil)
 	clusterState := clusterstate.NewClusterStateRegistry(provider, clusterstate.ClusterStateRegistryConfig{}, context.LogRecorder, newBackoff())
 	clusterState.UpdateNodes(nodes, nodeInfos, time.Now())
 
@@ -712,8 +708,8 @@ func TestScaleUpBalanceGroups(t *testing.T) {
 		pods = append(pods, BuildTestPod(fmt.Sprintf("test-pod-%v", i), 80, 0))
 	}
 
-	processors := ca_processors.TestProcessors()
-	scaleUpStatus, typedErr := ScaleUp(&context, processors, clusterState, pods, nodes, []*appsv1.DaemonSet{}, nodeInfos)
+	processors := NewTestProcessors()
+	scaleUpStatus, typedErr := ScaleUp(&context, processors, clusterState, pods, nodes, []*appsv1.DaemonSet{}, nodeInfos, nil)
 
 	assert.NoError(t, typedErr)
 	assert.True(t, scaleUpStatus.WasSuccessful())
@@ -759,22 +755,24 @@ func TestScaleUpAutoprovisionedNodeGroup(t *testing.T) {
 		NodeAutoprovisioningEnabled:      true,
 		MaxAutoprovisionedNodeGroupCount: 10,
 	}
-	context := NewScaleTestAutoscalingContext(options, fakeClient, nil, provider)
+	podLister := kube_util.NewTestPodLister([]*apiv1.Pod{})
+	listers := kube_util.NewListerRegistry(nil, nil, podLister, nil, nil, nil, nil, nil, nil, nil)
+	context := NewScaleTestAutoscalingContext(options, fakeClient, listers, provider, nil)
 
 	clusterState := clusterstate.NewClusterStateRegistry(provider, clusterstate.ClusterStateRegistryConfig{}, context.LogRecorder, newBackoff())
 
-	processors := ca_processors.TestProcessors()
+	processors := NewTestProcessors()
 	processors.NodeGroupListProcessor = &mockAutoprovisioningNodeGroupListProcessor{t}
 	processors.NodeGroupManager = &mockAutoprovisioningNodeGroupManager{t}
 
 	nodes := []*apiv1.Node{}
-	nodeInfos, _ := getNodeInfosForGroups(nodes, nil, provider, context.ListerRegistry, []*appsv1.DaemonSet{}, context.PredicateChecker)
+	nodeInfos, _ := utils.GetNodeInfosForGroups(nodes, nil, provider, context.ListerRegistry, []*appsv1.DaemonSet{}, context.PredicateChecker, nil)
 
-	scaleUpStatus, err := ScaleUp(&context, processors, clusterState, []*apiv1.Pod{p1}, nodes, []*appsv1.DaemonSet{}, nodeInfos)
+	scaleUpStatus, err := ScaleUp(&context, processors, clusterState, []*apiv1.Pod{p1}, nodes, []*appsv1.DaemonSet{}, nodeInfos, nil)
 	assert.NoError(t, err)
 	assert.True(t, scaleUpStatus.WasSuccessful())
-	assert.Equal(t, "autoprovisioned-T1", getStringFromChan(createdGroups))
-	assert.Equal(t, "autoprovisioned-T1-1", getStringFromChan(expandedGroups))
+	assert.Equal(t, "autoprovisioned-T1", utils.GetStringFromChan(createdGroups))
+	assert.Equal(t, "autoprovisioned-T1-1", utils.GetStringFromChan(expandedGroups))
 }
 
 func TestCheckScaleUpDeltaWithinLimits(t *testing.T) {
